searchData={"items":[{"type":"module","title":"Composite","doc":"A utility for writing dynamic queries.\n\nIt allows getting rid of some boilerplate when building a query based on input parameters.\n\n    params = %{search_query: \"John Doe\"}\n\n    User\n    |> where(active: true)\n    |> Composite.new(params)\n    |> Composite.param(:org_id, &filter_by_org_id/2)\n    |> Composite.param(:search_query, &search_by_full_name/2)\n    |> Composite.param(:org_name, &filter_by_org_name/2, requires: :org)\n    |> Composite.param(:org_type, &filter_by_org_type/2, requires: :org)\n    |> Composite.dependency(:org, &join_orgs/1)\n    |> Repo.all()\n\nEven though most of the examples in this doc use `Ecto`, Composite itself is not limited only to it.\n`Ecto` is an optional dependency and it is present only for having an implementation of `Ecto.Queryable` OOTB.\nYou're able to use Composite with any Elixir term, as it is just an advanced wrapper around `Enum.reduce/3`.","ref":"Composite.html"},{"type":"function","title":"Composite.apply/1","doc":"Applies handlers to query.\n\nUsed when composite is defined with `new/2`.\n\nIf used with `Ecto`, then calling this function is not necessary,\nas `Composite` implements `Ecto.Queryable` protocol, so applying will be done automatically when it is needed.","ref":"Composite.html#apply/1"},{"type":"function","title":"Composite.apply/3","doc":"Applies handlers to query.\n\nUsed when composite is defined with `new/0`","ref":"Composite.html#apply/3"},{"type":"function","title":"Composite.dependency/4","doc":"Defines a dependency loader.\n\nDependency is an instruction which is being applied lazily to a query.\nThe same dependency can be required by many parameters, but it will be invoked only once.\nDependency can depend on other dependency.\n\nUseful for joining tables.\n\n    params = %{org_type: :nonprofit, is_org_closed: false}\n\n    User\n    |> Composite.new(params)\n    |> Composite.param(:is_org_closed, &where(&1, [orgs: orgs], orgs.closed == ^&2), requires: :orgs)\n    |> Composite.param(:org_type, &where(&1, [orgs: orgs], orgs.type == ^&2), requires: :orgs)\n    |> Composite.dependency(:orgs, &join(&1, :inner, [users], orgs in assoc(users, :org), as: :orgs))\n\nIt is also possible to require a dependency only if specific value is set. In example below dependency `:phone` will be\nloaded only if value of `:search` param starts from `+` sign\n\n    composite\n    |> Composite.param(\n      :search,\n      fn\n        query, \"+\" <> _ = phone_number -> where(query, [phones: phones], phones.number == ^phone_number)\n        query, query_string -> where(query, [records], ilike(records.text, ^query_string))\n      end,\n      requires: fn\n        \"+\" <> _ -> :phone\n        _ -> nil\n      end\n    )\n    |> Composite.dependency(:phone, &join(&1, :inner, [records], phones in assoc(records, :phone), as: :phones))\n\nWhen `loader` function has arity 2, then all parameters are passed in the second argument.\n\n#","ref":"Composite.html#dependency/4"},{"type":"function","title":"Options - Composite.dependency/4","doc":"* `:requires` - allows to set dependencies for current dependency.","ref":"Composite.html#dependency/4-options"},{"type":"function","title":"Composite.force_require/2","doc":"Forces loading dependency even if it is not required by `params`.","ref":"Composite.html#force_require/2"},{"type":"function","title":"Composite.new/0","doc":"Initializes a `Composite` struct with delayed application of `query` and `params`.\n\nMust be used with `apply/3`.\n\n    composite =\n      Composite.new()\n      |> Composite.param(:organization_id, &where(&1, organization_id: ^&2))\n      |> Composite.param(:age_more_than, &where(&1, [users], users.age > ^&2))\n\n    params = %{organization_id: 1}\n\n    User\n    |> where(active: true)\n    |> Composite.apply(composite, params)\n    |> Repo.all()","ref":"Composite.html#new/0"},{"type":"function","title":"Composite.new/2","doc":"Initializes a `Composite` struct with `query` and `params`.\n\nMust be used with `apply/1`.\n\nThis strategy is useful when working with `Ecto.Query` in pipe-based queries.\n\n    params = %{organization_id: 1}\n\n    User\n    |> where(active: true)\n    |> Composite.new(params)\n    |> Composite.param(:organization_id, &where(&1, organization_id: ^&2))\n    |> Composite.param(:age_more_than, &where(&1, [users], users.age > ^&2))\n    |> Repo.all()\n\nPlease note, that there is no explicit `Composite.apply/1` call before `Repo.all/1`, because `Composite`\nimplements `Ecto.Queryable` protocol.","ref":"Composite.html#new/2"},{"type":"function","title":"Composite.param/4","doc":"Defines a parameter handler.\n\nHandler is applied to a query when `apply/1` or `apply/3` is invoked.\nAll handlers are invoke in the same order as they are defined.\n\nIf the parameter requires dependencies, then they will be loaded before the parameters' handler and only if\nparameter wasn't ignored. Examples with dependencies usage can be found in doc for `dependency/4`\n\n    params = %{location: \"Arctic\", order: :age_desc}\n\n    User\n    |> Composite.new(params)\n    |> Composite.param(:location, &where(&1, location: ^&2),\n      ignore?: &(&1 in [nil, \"WORLDWIDE\", \"\"])\n    )\n    |> Composite.param(\n      :order,\n      fn\n        query, :name_asc -> query |> order_by(name: :asc)\n        query, :age_desc -> query |> order_by(age: :desc)\n      end,\n      on_ignore: &order_by(&1, inserted_at: :desc)\n    )\n\nIf input parameters have nested maps (or any other key-based data structure):\n\n    params = %{filter: %{name: \"John\"}}\n\n    User\n    |> Composite.new(params)\n    |> Composite.param([:filter, :name], &where(&1, name: ^&2))\n\n#","ref":"Composite.html#param/4"},{"type":"function","title":"Options - Composite.param/4","doc":"* `:ignore?` - if function returns `true`, then handler `t:apply_fun/1` won't be applied.\nDefault value is `&(&1 in [nil, \"\", [], %{}])`.\n* `:on_ignore` - a function that will be applied instead of `t:apply_fun/1` if value is ignored.\nDefaults to `Function.identity/1`.\n* `:requires` - points to the dependencies which has to be loaded before calling `t:apply_fun/1`.\nIt is, also, possible to specify dependencies dynamically based on a value of the parameter by\npassing a function. The latter function will always receive not ignored values.\nDefaults to `nil` (which is equivalent to `[]`).\n* `:ignore_requires` - points to the dependencies which has to be loaded when value is ignored. May be needed\nfor custom `:on_ignore` implementation.","ref":"Composite.html#param/4-options"},{"type":"type","title":"Composite.apply_fun/1","doc":"","ref":"Composite.html#t:apply_fun/1"},{"type":"type","title":"Composite.dependencies/0","doc":"","ref":"Composite.html#t:dependencies/0"},{"type":"type","title":"Composite.dependency_name/0","doc":"","ref":"Composite.html#t:dependency_name/0"},{"type":"type","title":"Composite.dependency_option/0","doc":"","ref":"Composite.html#t:dependency_option/0"},{"type":"type","title":"Composite.load_dependency/1","doc":"","ref":"Composite.html#t:load_dependency/1"},{"type":"type","title":"Composite.param_option/1","doc":"","ref":"Composite.html#t:param_option/1"},{"type":"type","title":"Composite.param_path_item/0","doc":"","ref":"Composite.html#t:param_path_item/0"},{"type":"type","title":"Composite.params/0","doc":"","ref":"Composite.html#t:params/0"},{"type":"type","title":"Composite.t/1","doc":"","ref":"Composite.html#t:t/1"},{"type":"extras","title":"README","doc":"# Composite","ref":"readme.html"},{"type":"extras","title":"Installation - README","doc":"The package can be installed from [hex.pm](https://hex.pm/packages/composite) by\nadding `composite` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:composite, \"~> 0.4\"}\n  ]\nend\n```\nDocs can be found at [https://hexdocs.pm/composite](https://hexdocs.pm/composite).","ref":"readme.html#installation"},{"type":"extras","title":"About - README","doc":"Composite is a versatile utility library for building dynamic queries in Elixir. It simplifies the process of\nconstructing complex queries based on input parameters, making your code more concise and readable. While the library\nwas with Ecto in mind, it can be used with any Elixir term, as it's essentially an advanced wrapper around Enum.reduce/3.\n\nThe majority of the features of the library can be expressed by this example:\n```elixir\ndef list_users(params) do\n  MyApp.User\n  |> where(active: true)\n  |> Composite.new(params)\n  |> Composite.param(:name, &where(&1, name: ^&2))\n  |> Composite.param([:company, :name], &where(&1, [companies: companies], companies.name == ^&2),\n    requires: :companies\n  )\n  |> Composite.param(\n    :locations,\n    &where(&1, [departments: departments], departments.location in ^&2),\n    requires: :departments,\n    ignore?: &(&1 in [nil, []] or \"Worldwide\" in &1)\n  )\n  |> Composite.param(\n    :order,\n    fn\n      query, :department_name_asc ->\n        query |> order_by([departments: departments], asc: departments.name)\n\n      query, :username_asc ->\n        query |> order_by(asc: :name)\n    end,\n    requires: fn\n      :department_name_asc -> :departments\n      _ -> nil\n    end\n  )\n  |> Composite.dependency(:departments, fn query ->\n    join(query, :inner, [users], assoc(users, :department), as: :departments)\n  end)\n  |> Composite.dependency(\n    :companies,\n    fn query ->\n      join(query, :inner, [departments: departments], assoc(departments, :company), as: :companies)\n    end,\n    requires: :departments\n  )\n  |> MyApp.Repo.all()\nend\n```\n\nLet's move anonymous functions to named functions, so it doesn't look so scary:\n```elixir\ndef list_users(params) do\n  MyApp.User\n  |> where(active: true)\n  |> Composite.new(params)\n  |> Composite.param(:name, &where(&1, name: ^&2))\n  |> Composite.param([:company, :name], &filter_users_by_company_name/2, requires: :companies)\n  |> Composite.param(:locations, &filter_users_by_department_locations/2, requires: :departments, ignore?: &(&1 in [nil, []] or \"Worldwide\" in &1))\n  |> Composite.param(:order, &order_users/2, requires: &if(&1 == :department_name_asc, do: :departments))\n  |> Composite.dependency(:departments, &join_departments/1)\n  |> Composite.dependency(:companies, &join_companies/1, requires: :departments)\n  |> MyApp.Repo.all()\nend\n```\nYes, it looks like a router.\nThe example above starts with Ecto query. \nThe query is wrapped into `Composite` struct by calling `Composite.new/2`. It will be unwrapped automatically during\n`MyApp.Repo.all/1` call, as `Composite` implements `Ecto.Queryable` protocol.\nParameter handlers are defined using `Composite.param/3` and `Composite.param/4`. These instructions define\nhow the query should be modified if given parameters are present. By default parameters are ignored if they have\none of the following values: `nil`, `\"\"`, `[]`, `%{}`. However, this behaviour can be adjusted by using `:ignore?` option.\n\nParameter keys are not limited to atoms only, they can have any type. Lists have special meaning: if list is specified,\nthen this is a path to nested structure.\nHere is an example with all parameters for our `list_users/1` function:\n```elixir\n# all keys are optional\n%{name: \"John\", company: %{name: \"GitHub\"}, order: :department_name_asc, locations: [\"USA\"]}\n```\n\nAs you may noticed, there is a concept of dependencies.\nDependencies can be declared with `Composite.dependency/3` and `Composite.dependency/4` functions by specifying loader function.\nThe loader function is invoked before invoking parameter handler.\nDependencies can have other dependencies as well. Dependencies are loaded only when they're needed and only once.\nSo, if multiple parameter handlers require the same table to be joined to the query, it will be joined only once without any errors.","ref":"readme.html#about"}],"content_type":"text/markdown"}